from xently.core.loading import _find_registered_app_name, _import_module, _pluck_classes


def class_loader(module_label, classnames, module_prefix):
    """
    Dynamically import a list of classes from the given module.

    This works by looking up a matching app from the app registry,
    against the passed module label.  If the requested class can't be found in
    the matching module, then we attempt to import it from the corresponding
    xauth app, and if that fails, then the corresponding Oscar core app.

    Args:
        module_label (str): Module label comprising the app label and the
            module name, separated by a dot.  For example, 'catalogue.forms'.
        classname (str): Name of the class to be imported.

    Returns:
        The requested class object or ``None`` if it can't be found

    Raises:

        AppNotFoundError: If no app is found in ``INSTALLED_APPS`` that matches
            the passed module label.

        ImportError: If the attempted import of a class raises an
            ``ImportError``, it is re-raised
    """

    if "." not in module_label:
        # Importing from top-level modules is not supported, e.g.
        raise ValueError("Importing from top-level modules is not supported")

    # First look in xauth
    xauth_module = _import_module(f"xauth.{module_label}", classnames)
    # if nothing is there, then look in Oscar itself
    # e.g. 'oscar.apps.dashboard.catalogue.forms'
    oscar_module = _import_module(f"{module_prefix}.{module_label}", classnames)

    # returns e.g. 'oscar.apps.dashboard.catalogue',
    # 'yourproject.apps.dashboard.catalogue' or 'dashboard.catalogue',
    # depending on what is set in INSTALLED_APPS
    app_name = _find_registered_app_name(module_label)
    if app_name.startswith(f"{module_prefix}."):
        # The entry is obviously an Oscar one, we don't import again
        local_module = None
    else:
        # Attempt to import the classes from the local module
        # e.g. 'yourproject.dashboard.catalogue.forms'
        local_module_label = ".".join(app_name.split(".") + module_label.split(".")[1:])
        local_module = _import_module(local_module_label, classnames)

    if xauth_module is oscar_module is local_module is None:
        # This intentionally doesn't raise an ImportError, because ImportError
        # can get masked in complex circular import scenarios.
        raise ModuleNotFoundError(
            f"The module with label '{module_label}' could not be imported. This either"
            "means that it indeed does not exist, or you might have a problem"
            " with a circular import."
        )

    # return imported classes, giving preference to ones from the local package
    return _pluck_classes([local_module, xauth_module, oscar_module], classnames)



from django.apps import apps
from django.conf import settings
from django.contrib.auth.hashers import make_password, check_password
from django.contrib.auth.models import AbstractBaseUser
from django.contrib.auth.models import PermissionsMixin
from django.contrib.auth.password_validation import validate_password
from django.core import signing
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from xently.core.loading import get_class

from xauth.accounts import signing_salt
from xauth.internal_settings import (
    ENFORCE_ACCOUNT_VERIFICATION,
    AUTH_APP_LABEL,
    ACCESS_TOKEN_EXPIRY,
    VERIFICATION_CODE_EXPIRY,
    TEMPORARY_PASSWORD_EXPIRY,
    PASSWORD_RESET_REQUEST_SUBJECT,
    VERIFICATION_REQUEST_SUBJECT,
)

Token = get_class(f"{AUTH_APP_LABEL}.token.generator", "Token")
UserManager = get_class(f"{AUTH_APP_LABEL}.managers", "UserManager")
Mail = get_class(f"{AUTH_APP_LABEL}.mail", "Mail")

__all__ = [
    "AbstractUser",
    "AbstractSecurity",
    "AbstractSecurityQuestion",
    "AbstractFailedSignInAttempt",
    "AbstractPasswordResetLog",
    "default_is_verified",
]


def default_is_verified():
    return not ENFORCE_ACCOUNT_VERIFICATION


class AbstractUser(AbstractBaseUser, PermissionsMixin):
    email = models.EmailField(db_index=True, max_length=150, blank=False, unique=True)
    is_staff = models.BooleanField(default=False)
    is_verified = models.BooleanField(default=default_is_verified)

    objects = UserManager()

    USERNAME_FIELD = EMAIL_FIELD = "email"  # returned by get_email_field_name()

    # all the fields listed here(including the USERNAME_FIELD and password) are
    # expected as part of parameters in `objects`(UserManager).create_superuser
    # REQUIRED_FIELDS = ["email"]

    # Contains a tuple of fields that are "safe" to access publicly with proper
    # caution taken for modification
    READ_ONLY_FIELDS = ("is_superuser", "is_staff", "is_verified")

    WRITE_ONLY_FIELDS = ("password",)

    VERIFICATION_CODE_LENGTH = 6

    TEMPORARY_PASSWORD_LENGTH = 8

    _PASSWORD_RESET_REQUEST_FLAG_ATTR = "requested_password_reset"

    class Meta:
        abstract = True
        app_label = AUTH_APP_LABEL

    @classmethod
    def serializable_fields(cls):
        return ("email",) + cls.WRITE_ONLY_FIELDS + cls.READ_ONLY_FIELDS

    @classmethod
    def admin_panel_fields(cls):
        return ("email",)

    @property
    def token(self):
        if self.is_verified:
            if hasattr(self, self.__class__._PASSWORD_RESET_REQUEST_FLAG_ATTR):
                return self.password_reset_token
            return Token(self.token_payload, expiry_period=ACCESS_TOKEN_EXPIRY)
        return self.verification_token

    @property
    def signed_id(self):
        return signing.Signer(salt=signing_salt).sign(self.pk)

    @classmethod
    def from_signed_id(cls, signed_id):
        try:
            unsigned_id = signing.Signer(salt=signing_salt).unsign(signed_id)
        except signing.BadSignature:
            pass
        else:
            return cls.objects.get(pk=unsigned_id)

    @property
    def verification_token(self):
        return Token(self.token_payload, expiry_period=VERIFICATION_CODE_EXPIRY, subject="verification")

    @property
    def password_reset_token(self):
        return Token(self.token_payload, expiry_period=TEMPORARY_PASSWORD_EXPIRY, subject="password-reset")

    def _flag_password_reset(self):
        assert not hasattr(self, self.__class__._PASSWORD_RESET_REQUEST_FLAG_ATTR), "Cannot modify an existing attribute"
        setattr(self, self.__class__._PASSWORD_RESET_REQUEST_FLAG_ATTR, True)

    def unflag_password_reset(self):
        try:
            delattr(self, self.__class__._PASSWORD_RESET_REQUEST_FLAG_ATTR)
        except AttributeError:
            return False
        return True

    def request_password_reset(self, **kwargs):
        password = self.__class__.objects.make_random_password(self.__class__.TEMPORARY_PASSWORD_LENGTH)

        apps.get_model(AUTH_APP_LABEL, "Security").objects.update_or_create(
            user=self,
            defaults={"temporary_password": make_password(password), "temporary_password_generation_time": timezone.now},
        )

        self._flag_password_reset()

        if kwargs.copy().pop("send_email", False):
            kwargs.setdefault("subject", PASSWORD_RESET_REQUEST_SUBJECT)
            self._send_email("email-request-password-reset", {"password": password}, **kwargs)
        return password

    request_password_reset.alters_data = True

    def request_verification(self, **kwargs):
        if self.is_verified:
            return

        code = self.__class__.objects.make_random_password(self.__class__.VERIFICATION_CODE_LENGTH, "23456789")

        apps.get_model(AUTH_APP_LABEL, "Security").objects.update_or_create(
            user=self,
            defaults={"verification_code": make_password(code), "verification_code_generation_time": timezone.now},
        )

        if kwargs.copy().pop("send_email", False):
            kwargs.setdefault("subject", VERIFICATION_REQUEST_SUBJECT)
            self._send_email("email-request-verification", {"code": code}, **kwargs)
        return code

    request_verification.alters_data = True

    def set_password(self, raw_password):
        validate_password(raw_password, user=self)
        super().set_password(raw_password)

    def reset_password(self, old_password, new_password, is_change=False):
        try:
            matched = check_password(old_password, self.password if is_change else self.security.temporary_password)
        except ObjectDoesNotExist:
            return False
        else:
            if matched:
                self.set_password(new_password)
                self.save(update_fields=["password"])
            return matched

    reset_password.alters_data = True

    def verify(self, code):
        try:
            matched = check_password(code, self.security.verification_code)
        except ObjectDoesNotExist:
            return False
        else:
            if matched:
                self.is_verified = True
                self.save(update_fields=["is_verified"])
            return matched

    verify.alters_data = True

    def add_security_question(self, security_question, security_question_answer):
        encrypted_answer = make_password(security_question_answer)
        apps.get_model(AUTH_APP_LABEL, "Security").objects.update_or_create(
            user=self,
            defaults={
                "security_question": security_question,
                "security_question_answer": encrypted_answer,
            },
        )

    add_security_question.alters_data = True

    def _send_email(self, template_name, context=None, subject=None, **kwargs):
        context = context if context else {}
        context["user"] = self
        Mail(subject=subject).add_recipient(self.email).send(template_name=template_name, context=context, **kwargs)

    @cached_property
    def token_payload(self):
        return {"id": self.signed_id}


class AbstractSecurityQuestion(models.Model):
    question = models.CharField(max_length=255, blank=False, null=False, unique=True)
    added_on = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True
        app_label = AUTH_APP_LABEL

    def __str__(self):
        return self.question


class AbstractSecurity(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, related_name="security", on_delete=models.CASCADE, primary_key=True
    )
    security_question = models.ForeignKey("accounts.SecurityQuestion", on_delete=models.SET_NULL, null=True, blank=True)
    security_question_answer = models.CharField(max_length=150, null=True)
    temporary_password = models.CharField(max_length=128, blank=False, null=True)
    verification_code = models.CharField(max_length=128, blank=False, null=True)
    temporary_password_generation_time = models.DateTimeField(blank=True, null=True)
    verification_code_generation_time = models.DateTimeField(blank=True, null=True)
    account_deactivation_time = models.DateTimeField(blank=True, null=True)

    class Meta:
        abstract = True
        app_label = AUTH_APP_LABEL
        unique_together = ("user", "security_question")


class AbstractPasswordResetLog(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="password_reset", on_delete=models.CASCADE)
    type = models.CharField(choices=[("Reset", 1), ("Change", 2)], max_length=10, default=2)
    request_ip = models.GenericIPAddressField(db_index=True, unpack_ipv4=True, blank=True, null=True)
    change_ip = models.GenericIPAddressField(db_index=True, unpack_ipv4=True, blank=True, null=True)
    request_time = models.DateTimeField(default=timezone.now)
    change_time = models.DateTimeField(blank=False, null=True)

    class Meta:
        abstract = True
        app_label = AUTH_APP_LABEL


class AbstractFailedSignInAttempt(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="signin_attempts", on_delete=models.CASCADE)
    device_ip = models.GenericIPAddressField(db_index=True, unpack_ipv4=True, blank=True, null=True)
    attempt_date = models.DateField(default=timezone.now)
    attempt_count = models.IntegerField(default=1)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        app_label = AUTH_APP_LABEL



from django.apps import apps
from django.contrib.auth import get_user_model, login, logout
from django.utils.translation import gettext as _
from rest_framework import viewsets, permissions, exceptions
from rest_framework.decorators import action
from rest_framework.response import Response
from xently.core.loading import get_classes, get_class

from xauth.internal_settings import AUTH_APP_LABEL

IsOwner = get_class(f"{AUTH_APP_LABEL}.permissions", "IsOwner")
(
    SecurityQuestionSerializer,
    ProfileSerializer,
    AccountVerificationSerializer,
    PasswordResetSerializer,
    AddSecurityQuestionSerializer,
    AccountActivationSerializer,
) = get_classes(
    f"{AUTH_APP_LABEL}.serializers",
    [
        "SecurityQuestionSerializer",
        "ProfileSerializer",
        "AccountVerificationSerializer",
        "PasswordResetSerializer",
        "AddSecurityQuestionSerializer",
        "AccountActivationSerializer",
    ],
)

__all__ = ["AccountViewSet", "SecurityQuestionViewSet"]


class SecurityQuestionViewSet(viewsets.ModelViewSet):
    serializer_class = SecurityQuestionSerializer
    queryset = apps.get_model(AUTH_APP_LABEL, "SecurityQuestion").objects.all()
    permission_classes = [permissions.IsAdminUser]  # TODO: consider superuser only


class AccountViewSet(viewsets.ModelViewSet):
    serializer_class = ProfileSerializer
    permission_classes = [IsOwner]
    queryset = get_user_model().objects.all()

    def get_queryset(self):
        return get_user_model().objects.filter(pk=self.request.user.pk)

    def get_object(self):
        lookup_key = self.lookup_url_kwarg or self.lookup_field
        if lookup_key not in self.kwargs or self.action == "request_temporary_password":
            obj = self.request.user
            self.check_object_permissions(self.request, obj)
        else:
            obj = super().get_object()
        return obj

    def get_serializer_context(self):
        context = super().get_serializer_context().copy()
        remove_fields = []
        if self.action in ["update", "partial_update"]:
            # Password should not be updated using this action. Use `reset_password` instead
            remove_fields += ["password", "token"]
        elif self.action in ["set_security_question", "retrieve", "list"]:
            remove_fields.append("token")
        elif self.action == "signout":
            remove_fields = "__all__"
        context["remove_fields"] = remove_fields
        return context

    def retrieve(self, request, *args, **kwargs):
        # We need response to return data as per class declared serializer class
        self.serializer_class = self.__class__.serializer_class
        return super().retrieve(request, *args, **kwargs)

    @action(methods=["POST"], detail=False, permission_classes=[permissions.AllowAny])
    def signup(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    @action(methods=["POST"], detail=False)
    def signin(self, request, *args, **kwargs):
        login(request, request.user)
        return self.retrieve(request, *args, **kwargs)

    @action(methods=["POST"], detail=True)
    def signout(self, request, *args, **kwargs):
        logout(request)
        return Response()

    def do_request_verification_code(self, user):
        user.request_verification(send_mail=True, request=self.request)

    @action(detail=True, url_path="request-verification-code", serializer_class=None)
    def request_verification_code(self, request, *args, **kwargs):
        self.do_request_verification_code(request.user)
        return self.retrieve(request, *args, **kwargs)

    @action(methods=["POST"], detail=True, url_path="verify-account", serializer_class=AccountVerificationSerializer)
    def verify_account(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid(raise_exception=True) and self.request.user.verify(**serializer.validated_data):
            return self.retrieve(request, *args, **kwargs)
        raise exceptions.ValidationError(_("Invalid verification code."))

    def do_request_temporary_password(self, user):
        user.request_password_reset(send_mail=True, request=self.request)

    @action(detail=True, url_path="request-temporary-password", serializer_class=None)
    def request_temporary_password(self, request, *args, **kwargs):
        self.do_request_temporary_password(request.user)
        response = self.retrieve(request, *args, **kwargs)
        request.user.unflag_password_reset()
        return response

    @action(methods=["POST"], detail=True, url_path="reset-password", serializer_class=PasswordResetSerializer)
    def reset_password(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=self.request.data)
        if serializer.is_valid(raise_exception=True) and self.request.user.reset_password(**serializer.validated_data):
            return self.retrieve(request, *args, **kwargs)
        is_change = serializer.validated_data["is_change"]
        raise exceptions.ValidationError(_(f"Invalid {'old' if is_change else 'temporary'} password."))

    @action(
        methods=["POST"], detail=True, url_path="set-security-question", serializer_class=AddSecurityQuestionSerializer
    )
    def set_security_question(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            self.request.user.add_security_question(**serializer.validated_data)
            return Response()
        raise exceptions.ValidationError(_("Invalid security question"))

    @action(methods=["POST"], detail=True, url_path="activate-account", serializer_class=AccountActivationSerializer)
    def activate_account(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        try:
            if serializer.is_valid(raise_exception=True) and request.user.activate_account(**serializer.validated_data):
                return self.retrieve(request, *args, **kwargs)
        except AttributeError:
            raise AttributeError("Make sure you applied `models.UserActivationMixin` to your `settings.AUTH_USER_MODEL`")
        return exceptions.ValidationError(_("Wrong security question answer"))



import json

from django.utils.datetime_safe import datetime
from jwcrypto import jwt
from xently.core.loading import get_class

from xauth.internal_settings import TOKEN_EXPIRY, REQUEST_TOKEN_ENCRYPTED, AUTH_APP_LABEL

__all__ = ["Token"]


class Token(get_class(f"{AUTH_APP_LABEL}.token.key", "TokenKey")):
    def __init__(self, payload, activation_date=None, expiry_period=None, payload_key=None, subject=None):
        super().__init__()
        self._unencrypted = None
        self._encrypted = None
        self.subject = subject or "access"
        self.payload = payload
        self.payload_key = payload_key or "payload"
        self.activation_date = activation_date
        self.expiry_period = expiry_period

    def __repr__(self):
        return self.encrypted

    @property
    def unencrypted(self):
        if self._unencrypted is None:
            self.refresh()
        return self._unencrypted

    @property
    def encrypted(self):
        if self._encrypted is None:
            self.refresh()
        return self._encrypted

    @property
    def checked_claims(self):
        issue_date = datetime.now()
        self.activation_date = issue_date if not self.activation_date else self.activation_date
        self.expiry_period = TOKEN_EXPIRY if self.expiry_period is None else self.expiry_period

        expiry_date = self.activation_date + self.expiry_period
        activation_secs = int(self.activation_date.strftime("%s"))
        expiry_secs = int(expiry_date.strftime("%s"))
        assert expiry_secs > activation_secs, "Expiration date must be a date later than activation date"
        return {
            "nbf": activation_secs,
            "exp": expiry_secs,
            "iat": int(issue_date.strftime("%s")),
            "sub": self.subject,
        }

    @property
    def claims(self):
        # convert provided payload to string if it is not already a string or a dictionary
        # self.payload = str(self.payload) if not isinstance(self.payload, dict) else self.payload
        cc = self.checked_claims
        cc[self.payload_key] = self.payload
        return cc

    @property
    def tokens(self):
        if not self.unencrypted or not self.encrypted:
            self.refresh()
        return {"unencrypted": self.unencrypted, "encrypted": self.encrypted}

    def get_claims(self, token=None, encrypted: bool = REQUEST_TOKEN_ENCRYPTED):
        token = self.encrypted if not token else token
        assert token is not None, "Call .refresh() first or provide a token"
        token = token.decode() if isinstance(token, bytes) else token
        try:
            token = jwt.JWT(key=self.encryption_key, jwt=f"{token}").claims if encrypted else token
        except ValueError:
            raise jwt.JWException
        return json.loads(jwt.JWT(key=self.get_jwt_signing_keys()[1], jwt=token).claims)

    def get_payload(self, token=None, encrypted: bool = REQUEST_TOKEN_ENCRYPTED):
        try:
            return self.get_claims(token, encrypted).get(self.payload_key, None)
        except AssertionError:
            return self.payload

    def refresh(self):
        header = {"alg": self.signing_algorithm, "typ": "JWT"}
        # unencrypted token
        token = jwt.JWT(header, self.claims, check_claims=self.checked_claims, algs=self.ALLOWED_SIGNING_ALGORITHMS)
        token.make_signed_token(key=self.get_jwt_signing_keys()[0])
        self._unencrypted = token.serialize()
        header = {"alg": "ECDH-ES", "enc": "A256GCM"}
        # encrypted token
        e_token = jwt.JWT(header=header, claims=self.unencrypted)
        e_token.make_encrypted_token(key=self.encryption_key)
        self._encrypted = e_token.serialize()
        return self.tokens



from django.apps import apps


def is_model_registered(app_label, model_name):
    try:
        apps.get_registered_model(app_label, model_name)
    except LookupError:
        return False
    else:
        return True



from xently.core.loading import is_model_registered

from xauth.accounts.abstract_models import (
    AbstractUser,
    AbstractSecurityQuestion,
    AbstractSecurity,
    AbstractPasswordResetLog,
    AbstractFailedSignInAttempt,
)
from xauth.internal_settings import AUTH_APP_LABEL

__all__ = []

if not is_model_registered("xauth", "User"):

    class User(AbstractUser):
        class Meta(AbstractUser.Meta):
            swappable = "AUTH_USER_MODEL"
            app_label = "xauth"

    __all__.append("User")

if not is_model_registered(AUTH_APP_LABEL, "SecurityQuestion"):

    class SecurityQuestion(AbstractSecurityQuestion):
        pass

    __all__.append("SecurityQuestion")

if not is_model_registered(AUTH_APP_LABEL, "Security"):

    class Security(AbstractSecurity):
        pass

    __all__.append("Security")

if not is_model_registered(AUTH_APP_LABEL, "PasswordResetLog"):

    class PasswordResetLog(AbstractPasswordResetLog):
        pass

    __all__.append("PasswordResetLog")

if not is_model_registered(AUTH_APP_LABEL, "FailedSignInAttempt"):

    class FailedSignInAttempt(AbstractFailedSignInAttempt):
        pass

    __all__.append("FailedSignInAttempt")



from django.contrib.auth.hashers import check_password
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from xently.core.loading import get_class

from xauth.internal_settings import ACCOUNT_ACTIVATION_TOKEN_EXPIRY, AUTH_APP_LABEL

Token = get_class(f"{AUTH_APP_LABEL}.token.generator", "Token")

__all__ = ["UserActivationMixin", "NameMixin"]


class UserActivationMixin(models.Model):
    is_active = models.BooleanField(default=True)

    class Meta:
        abstract = True

    @property
    def activation_token(self):
        return Token(self.token_payload, expiry_period=ACCOUNT_ACTIVATION_TOKEN_EXPIRY, subject="activation")

    @property
    def token(self):
        if not self.is_active:
            return self.activation_token
        return super().token

    def activate_account(self, security_question_answer):
        if self.is_active:
            return

        try:
            matched = check_password(security_question_answer, self.security.security_question_answer)
        except ObjectDoesNotExist:
            return False
        else:
            if matched:
                self.is_active = True
                self.save(update_fields=["is_active"])
            return matched


class NameMixin(models.Model):
    first_name = models.CharField(db_index=True, max_length=50, blank=True, null=True)
    last_name = models.CharField(db_index=True, max_length=50, blank=True, null=True)

    class Meta:
        abstract = True

    def get_full_name(self):
        """
        This method is required by Django for things like handling emails.
        Typically this would be the user's first and last name. Since we do
        not store the user's real name, we return their username instead.
        """
        first_name = self.first_name.strip().capitalize() if self.first_name else ""
        last_name = self.last_name.strip().capitalize() if self.last_name else ""

        name = f"{first_name} {last_name}".strip()
        if not name and hasattr(self, "username"):
            # name is empty, use username instead
            name = self.username
        return name if name else None

    def get_short_name(self):
        """
        This method is required by Django for things like handling emails.
        Typically, this would be the user's first name. Since we do not store
        the user's real name, we return their username instead.
        """
        name = self.get_full_name()
        if name:
            return name.split()[0] if " " in name else name
        return name

    @property
    def name(self):
        return self.get_full_name()

    def __str__(self):
        return self.name



import os
from hashlib import md5
from pathlib import Path

from django.conf import settings
from jwcrypto import jwk
from jwcrypto.common import json_decode

from xauth.internal_settings import MAKE_KEY_DIRS, KEYS_DIR, JWT_SIG_ALG

__all__ = ["TokenKey"]


class TokenKey:
    ALLOWED_SIGNING_ALGORITHMS = ["RS256", "HS256"]

    def __init__(self, password=None, signing_algorithm=None):
        self.password = (password or settings.SECRET_KEY).encode()
        self.signing_algorithm = signing_algorithm or JWT_SIG_ALG
        assert (
            self.signing_algorithm in self.__class__.ALLOWED_SIGNING_ALGORITHMS
        ), f"{self.signing_algorithm} must be one of {self.__class__.ALLOWED_SIGNING_ALGORITHMS}"
        if MAKE_KEY_DIRS:
            Path(KEYS_DIR).mkdir(parents=True, exist_ok=True)

    def get_or_create_rs_256_key(self, file, key, is_private):
        password = self.password if is_private else None
        try:
            # get key from .pem file contents
            with open(file, "rb") as pem:
                key = jwk.JWK.from_pem(pem.read(), password=password)
        except FileNotFoundError:
            with open(file, "wb") as pem:
                # Write the key's to .pem file
                pem.write(key.export_to_pem(private_key=is_private, password=password))
        return key

    def get_or_create_hs_256_key(self, file, key):
        try:
            with open(file, "rb") as key_file:
                key = jwk.JWK(**json_decode(key_file.readline()))
        except FileNotFoundError:
            with open(file, "wb") as key_file:
                key_file.write(key.export().encode())
        return key

    def _get_jwt_signing_or_encryption_key(self, is_private=True, is_encryption=False):
        if is_encryption:
            file_name, key = "encryption_key", jwk.JWK.generate(kty="EC", alg="ECDH-ES", crv="P-256")
        else:
            file_name = "signing_key"
            key_op = "verify"  # `Public Key` will be used for `verifying` the `token`
            file_name += "_pub"  # `public` signing key `file name`
            if is_private:
                key_op = "sign"  # `Private Key` will be used for `signing` the `token`
                file_name += "_pri"  # `private` signing key `file name`
            key = jwk.JWK.generate(kty="RSA", key_ops=key_op, alg="RSA-OAEP", size=2048)

        file_name = md5(file_name.encode(encoding="utf8", errors="replace")).hexdigest()
        return self.get_or_create_rs_256_key(os.path.join(KEYS_DIR, f"{file_name}.pem"), key, is_private)

    @property
    def encryption_key(self):
        return self._get_jwt_signing_or_encryption_key(is_encryption=True)

    def get_jwt_signing_keys(self):
        """
        Return a `JWK` key for `JWT` token encryption or a `tuple` of JWK keys for `JWT` token signing
        (`private` key, `public` key) depending on the signing algorithm used
        :return: `JWK` key or `tuple` of `JWKs` keys
        """

        if self.signing_algorithm == "HS256":
            # Default signing algorithm used when all else fails.
            # `JWT` signing key
            jwt_sig_key = self.get_or_create_hs_256_key(
                Path(KEYS_DIR) / md5("signing_key".encode(encoding="utf8", errors="replace")).hexdigest(),
                jwk.JWK(generate="oct", size=256),
            )

            # Public `JWT` signing key for `JWT` verification
            jwt_pub_sig_key = jwt_sig_key
            # Private `JWT` signing key for `JWT` signing
            jwt_pri_sig_key = jwt_sig_key
        else:
            # Private `JWT` signing key for `JWT` signing
            jwt_pri_sig_key = self._get_jwt_signing_or_encryption_key()
            # Public `JWT` signing key for `JWT` verification
            jwt_pub_sig_key = self._get_jwt_signing_or_encryption_key(False)

        # Use the same key(private) for **signing** and **verifying** keys...
        return jwt_pri_sig_key, jwt_pri_sig_key or jwt_pub_sig_key



from django import forms
from django.apps import apps
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.forms import ReadOnlyPasswordHashField
from django.utils.translation import gettext as _

from xauth.internal_settings import AUTH_APP_LABEL

__all__ = ["UserCreationForm", "UserUpdateForm", "UserAdmin", "SecurityQuestionAdmin"]


class UserCreationForm(forms.ModelForm):
    """
    A form for creating new users. Includes all the required
    fields, plus a repeated password.
    """

    password1 = forms.CharField(label="Password", widget=forms.PasswordInput)
    password2 = forms.CharField(label="Password confirmation", widget=forms.PasswordInput)

    class Meta:
        model = get_user_model()
        fields = model.admin_panel_fields()

    def clean_password2(self):
        """check that two password entries match"""
        p1 = self.cleaned_data.get("password1")
        p2 = self.cleaned_data.get("password2")

        if p1 and p2 and p1 != p2:
            raise forms.ValidationError(_("Passwords do not match"), code="password_mismatch")

        return p2

    def save(self, commit=True):
        """save the user and provided password in hashed format"""
        user = super().save(commit=False)  # `commit=False` Helps avoid double database writes
        user.set_password(self.cleaned_data.get("password2"))

        if commit:
            user.save()

        return user


class UserUpdateForm(forms.ModelForm):
    """
    A form for updating users. Includes all the fields on
    the user, but replaces the password field with admin's
    password hash display field.
    """

    password = ReadOnlyPasswordHashField

    class Meta:
        model = get_user_model()
        fields = model.serializable_fields()

    def clean_password(self):
        # Regardless of what the user provides, return the initial value.
        # This is done here, rather than on the field, because the
        # field does not have access to the initial value
        return self.initial.get("password")


class UserAdmin(BaseUserAdmin):
    form = UserUpdateForm
    add_form = UserCreationForm

    ordering = (get_user_model().USERNAME_FIELD,)
    list_display = get_user_model().admin_panel_fields() + get_user_model().READ_ONLY_FIELDS
    list_filter = get_user_model().READ_ONLY_FIELDS
    fieldsets = [
        (None, {"fields": get_user_model().admin_panel_fields() + ("password",)}),
        ("Permissions", {"fields": get_user_model().READ_ONLY_FIELDS}),
    ]
    add_fieldsets = [
        (
            None,
            {
                "classes": ["wide"],  # Class for applying styling
                "fields": get_user_model().admin_panel_fields() + ("password1", "password2"),
            },
        )
    ]
    filter_horizontal = ()


class SecurityQuestionAdmin(admin.ModelAdmin):
    list_display = ("question", "added_on")


admin.site.register(get_user_model(), UserAdmin)
admin.site.register(apps.get_model(AUTH_APP_LABEL, "SecurityQuestion"), SecurityQuestionAdmin)
